package org.xbib.event.common;

import java.lang.reflect.InvocationTargetException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.StandardCopyOption;
import java.nio.file.attribute.FileTime;
import java.time.Instant;
import java.util.HashMap;
import java.util.Map;
import java.util.ServiceLoader;

import org.xbib.datastructures.json.tiny.Json;
import org.xbib.event.Event;
import org.xbib.event.EventConsumer;
import org.xbib.event.FileFollowEvent;
import org.xbib.event.Listener;
import org.xbib.event.Payload;
import org.xbib.event.bus.AsyncEventBus;
import org.xbib.event.bus.EventBus;
import org.xbib.event.bus.SubscriberExceptionContext;
import org.xbib.event.bus.SubscriberExceptionHandler;
import org.xbib.event.clock.ClockEventManagerService;
import org.xbib.event.generic.GenericEventManagerService;
import org.xbib.event.path.FileFollowEventManagerService;
import org.xbib.event.path.PathEventManagerService;
import org.xbib.event.timer.TimerEventManagerService;
import org.xbib.settings.Settings;

import java.io.Closeable;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.logging.Level;
import java.util.logging.Logger;

public final class EventManager {

    private static final Logger logger = Logger.getLogger(EventManager.class.getName());

    private final EventManagerBuilder builder;

    private static final Map<String, Class<? extends Event>> eventTypes = new HashMap<>();

    private final Map<Class<? extends EventManagerService>, EventManagerService> eventManagerServices;

    private EventManager(EventManagerBuilder builder) {
        this.builder = builder;
        eventTypes.put("null", NullEvent.class);
        eventTypes.put("generic", GenericEventImpl.class);
        eventTypes.put("path", PathEventImpl.class);
        eventTypes.put("clock", ClockEventImpl.class);
        eventTypes.put("timer", TimerEventImpl.class);
        eventTypes.put("filefollow", FileFollowEventImpl.class);
        for (Event event : ServiceLoader.load(Event.class)) {
            eventTypes.put(event.getType(), event.getClass());
        }
        logger.log(Level.INFO, "installed events = " + eventTypes.keySet());
        this.eventManagerServices = new HashMap<>();
        eventManagerServices.put(GenericEventManagerService.class, new GenericEventManagerService().init(this));
        eventManagerServices.put(ClockEventManagerService.class, new ClockEventManagerService().init(this));
        eventManagerServices.put(TimerEventManagerService.class, new TimerEventManagerService().init(this));
        eventManagerServices.put(FileFollowEventManagerService.class, new FileFollowEventManagerService().init(this));
        eventManagerServices.put(PathEventManagerService.class, new PathEventManagerService().init(this));
        for (EventManagerService service : ServiceLoader.load(EventManagerService.class)) {
            eventManagerServices.put(service.getClass(), service.init(this));
        }
        logger.log(Level.INFO, "installed event service managers = " + eventManagerServices.keySet());
    }

    public static EventManagerBuilder builder() {
        return new EventManagerBuilder();
    }

    public static EventBuilder eventBuilder() {
        return new EventBuilder();
    }

    public Settings getSettings() {
        return builder.settings;
    }

    public EventBus getEventBus() {
        return builder.eventBus;
    }

    public ClassLoader getClassLoader() {
        return builder.classLoader;
    }

    public ExecutorService getExecutorService() {
        return builder.executorService;
    }

    public void dispatch(Event event) {
        getGenericEventManagerService().post(event);
    }

    @SuppressWarnings("unchecked")
    public <T extends EventManagerService> T getEventManagerService(Class<T> cl) {
        return (T) eventManagerServices.get(cl);
    }

    public GenericEventManagerService getGenericEventManagerService() {
        return (GenericEventManagerService) eventManagerServices.get(GenericEventManagerService.class);
    }

    public ClockEventManagerService getClockEventManagerService() {
        return (ClockEventManagerService) eventManagerServices.get(ClockEventManagerService.class);
    }

    public TimerEventManagerService getTimerEventManagerService() {
        return (TimerEventManagerService) eventManagerServices.get(TimerEventManagerService.class);
    }

    public FileFollowEventManagerService getFileFollowEventManagerService() {
        return (FileFollowEventManagerService) eventManagerServices.get(FileFollowEventManagerService.class);
    }

    public PathEventManagerService getPathEventManagerService() {
        return (PathEventManagerService) eventManagerServices.get(PathEventManagerService.class);
    }

    public void close() throws IOException {
        for (EventConsumer eventConsumer : builder.eventConsumers) {
            if (eventConsumer instanceof Closeable closeable) {
                closeable.close();
            }
        }
        for (EventManagerService service : eventManagerServices.values()) {
            try {
                service.shutdown();
            } catch (IOException e) {
                logger.log(Level.SEVERE, e.getMessage(), e);
            }
        }
    }

    public static class EventManagerBuilder {

        private Settings settings;

        private int threadcount;

        private ExecutorService executorService;

        private ClassLoader classLoader;

        private SubscriberExceptionHandler subscriberExceptionHandler;

        private EventBus eventBus;

        private final List<EventConsumer> eventConsumers;

        private EventManagerBuilder() {
            this.eventConsumers = new ArrayList<>();
            this.subscriberExceptionHandler = new EventManagerExceptionHandler();
        }

        public EventManagerBuilder setSettings(Settings settings) {
            this.settings = settings;
            return this;
        }

        public EventManagerBuilder setThreadCount(int threadCount) {
            this.threadcount = threadCount;
            return this;
        }

        public EventManagerBuilder setExecutorService(ExecutorService executorService) {
            this.executorService = executorService;
            return this;
        }

        public EventManagerBuilder setClassLoader(ClassLoader classLoader) {
            this.classLoader = classLoader;
            return this;
        }

        public EventManagerBuilder setEventBus(EventBus eventBus) {
            this.eventBus = eventBus;
            return this;
        }

        public EventManagerBuilder register(String eventType, Class<? extends Event> eventClass) {
            eventTypes.put(eventType, eventClass);
            return this;
        }

        public EventManagerBuilder register(EventConsumer eventConsumer) {
            Objects.requireNonNull(eventConsumer, "event consumer must not be null");
            eventConsumers.add(eventConsumer);
            return this;
        }

        public EventManagerBuilder setSubscriberExceptionHandler(SubscriberExceptionHandler subscriberExceptionHandler) {
            this.subscriberExceptionHandler = subscriberExceptionHandler;
            return this;
        }

        public EventManagerBuilder loadEventConsumers() {
            if (classLoader == null) {
                classLoader = EventManager.class.getClassLoader();
            }
            for (EventConsumer eventConsumer : ServiceLoader.load(EventConsumer.class, classLoader)) {
                eventConsumers.add(eventConsumer);
            }
            return this;
        }

        public EventManager build() {
            if (settings == null) {
                settings = Settings.emptySettings();
            }
            if (executorService == null) {
                if (threadcount == 0) {
                    executorService = Executors.newCachedThreadPool();
                } else {
                    executorService = Executors.newFixedThreadPool(threadcount);
                }
            }
            if (classLoader == null) {
                classLoader = EventManager.class.getClassLoader();
            }
            if (eventBus == null) {
                eventBus = new AsyncEventBus(executorService, subscriberExceptionHandler);
            }
            for (EventConsumer eventConsumer : eventConsumers) {
                eventBus.register(eventConsumer);
            }
            return new EventManager(this);
        }
    }

    private static class EventManagerExceptionHandler implements SubscriberExceptionHandler {
        @Override
        public void handleException(Throwable exception, SubscriberExceptionContext context) {
            logger.log(Level.SEVERE, exception.getMessage(), exception);
        }
    }

    public static Event eventFromFile(Path file) throws IOException {
        return eventFromJson(Files.readString(file));
    }

    public static Event eventFromJson(String json) {
        Map<String, Object> map = Json.toMap(json);
        return eventBuilder()
                .setType(map.getOrDefault("type", "generic").toString())
                .setPayload(new Payload(map))
                .build();
    }

    public static class EventImpl implements Event {

        private final EventBuilder builder;

        EventImpl(EventBuilder builder) {
            this.builder = builder;
        }

        public Listener getListener() {
            return builder.listener;
        }

        @Override
        public String getType() {
            return builder.type;
        }

        @Override
        public String getCode() {
            return builder.code;
        }

        @Override
        public String getMessage() {
            return builder.message;
        }

        @Override
        public Payload getPayload() {
            return builder.payload;
        }

        @Override
        public Instant getCreated() {
            return builder.created;
        }

        @Override
        public Instant getScheduledFor() {
            return builder.scheduled;
        }

        @Override
        public Path getPath() {
            return builder.path;
        }

        @Override
        public String getBase() {
            return builder.base;
        }

        @Override
        public String getSuffix() {
            return builder.suffix;
        }

        @Override
        public long getFileSize() {
            return builder.fileSize;
        }

        public String toJson() throws IOException {
            return Json.toString(builder.payload);
        }

        @Override
        public boolean isNullEvent() {
            return builder.type == null;
        }

        @Override
        public String toString() {
            return "Event[path = " + builder.path + ", base = " + builder.base + ", suffix = " + builder.suffix + " payload = " + builder.payload +"]";
        }

        public void success() throws IOException {
            if (builder.path != null) {
                Files.setLastModifiedTime(builder.path, FileTime.from(Instant.now()));
                Files.move(builder.path, builder.path.getParent().resolve(Event.SUCCESS)
                                .resolve(builder.path.getFileName()).toAbsolutePath(),
                        StandardCopyOption.REPLACE_EXISTING);

            }
        }

        public void fail() throws IOException {
            Files.setLastModifiedTime(builder.path, FileTime.from(Instant.now()));
            Files.move(builder.path, builder.path.getParent().resolve(Event.FAIL)
                            .resolve(builder.path.getFileName()).toAbsolutePath(),
                    StandardCopyOption.REPLACE_EXISTING);
        }
    }

    public static class EventBuilder {

        Listener listener;

        String type;

        String code;

        String message;

        Instant scheduled;

        Instant created;

        Payload payload;

        Path path;

        String base;

        String suffix;

        long fileSize;

        long maxFileSize;

        EventBuilder() {
            this.maxFileSize = -1L;
        }

        public EventBuilder setListener(Listener listener) {
            this.listener = listener;
            return this;
        }

        public EventBuilder setType(String type) {
            if (type != null && eventTypes.containsKey(type)) {
                this.type = type;
            } else {
                throw new IllegalArgumentException("unregistered event type: " + type);
            }
            return this;
        }

        public EventBuilder setCode(String code) {
            this.code = code;
            return this;
        }

        public EventBuilder setMessage(String message) {
            this.message = message;
            return this;
        }

        public EventBuilder setScheduledFor(Instant scheduled) {
            this.scheduled = scheduled;
            return this;
        }

        public EventBuilder setPayload(Payload payload) {
            this.payload = payload;
            return this;
        }

        public EventBuilder setMaxFileSize(long maxFileSize) {
            this.maxFileSize = maxFileSize;
            return this;
        }

        public EventBuilder setPath(Path path) throws IOException {
            this.path = path;
            base = getBase(path);
            suffix = getSuffix(path);
            fileSize = Files.size(path);
            if (maxFileSize != -1L && fileSize > maxFileSize) {
                throw new IOException("file size too large");
            }
            return this;
        }

        public Event build() {
            if (type == null) {
                type = "null";
            }
            Class<? extends Event> cl = eventTypes.get(type);
            if (cl == null) {
                cl = NullEvent.class;
            }
            try {
                this.created = Instant.now();
                if (listener != null) {
                    return cl.getDeclaredConstructor(EventBuilder.class, Listener.class).newInstance(this, listener);
                } else {
                    return cl.getDeclaredConstructor(EventBuilder.class).newInstance(this);
                }
            } catch (NoSuchMethodException | InstantiationException | IllegalAccessException |
                     InvocationTargetException e) {
                logger.log(Level.WARNING, "unable to construct event object for type " + type + ", falling back to NullEvent");
                return new NullEvent(this);
            }
        }

        private static String getBase(Path path) {
            String name = path.getFileName().toString();
            int pos = name.lastIndexOf('.');
            return pos >= 0 ? name.substring(0, pos) : name;
        }

        private static String getSuffix(Path path) {
            String name = path.getFileName().toString();
            int pos = name.lastIndexOf('.');
            return pos >= 0 ? name.substring(pos + 1) : null;
        }
    }

}
