package org.xbib.event.path;

import org.xbib.datastructures.api.TimeValue;
import org.xbib.event.Event;
import org.xbib.event.bus.EventBus;
import org.xbib.event.common.EventManager;

import java.io.Closeable;
import java.io.IOException;
import java.io.UncheckedIOException;
import java.nio.file.ClosedWatchServiceException;
import java.nio.file.DirectoryStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.StandardWatchEventKinds;
import java.nio.file.WatchEvent;
import java.nio.file.WatchKey;
import java.nio.file.WatchService;
import java.nio.file.attribute.BasicFileAttributeView;
import java.nio.file.attribute.BasicFileAttributes;
import java.time.Instant;
import java.util.concurrent.Callable;
import java.util.logging.Level;
import java.util.logging.Logger;

public class PathEventService implements Callable<Integer>, Closeable {

    private static final Logger logger = Logger.getLogger(PathEventService.class.getName());


    private final PathEventManagerService pathEventManager;

    private final EventBus eventBus;

    private final Path path;

    private final String name;

    private final TimeValue lifetime;

    private final WatchService watchService;

    private int eventCount;

    private volatile boolean keepWatching;

    public PathEventService(PathEventManagerService pathEventManager,
                            EventBus eventBus,
                            String name,
                            Path path,
                            TimeValue lifetime) throws IOException {
        this.pathEventManager = pathEventManager;
        this.eventBus = eventBus;
        this.name = name;
        this.path = path;
        this.lifetime = lifetime;
        drainIncoming();
        this.watchService = path.getFileSystem().newWatchService();
        WatchEvent.Kind<?>[] kinds = new WatchEvent.Kind[] { StandardWatchEventKinds.ENTRY_CREATE };
        WatchKey watchKey = path.resolve(Event.INCOMING).register(watchService, kinds);
        keepWatching = true;
        logger.log(Level.INFO, "path event service created for incoming files at " + path);
    }

    public String getName() {
        return name;
    }


    @SuppressWarnings("unchecked")
    @Override
    public Integer call() {
        try {
            logger.log(Level.INFO, "watch service running on " + path.resolve(Event.INCOMING));
            while (keepWatching && watchService != null) {
                WatchKey watchKey = watchService.take();
                logger.log(Level.FINE, "received a watch key " + watchKey);
                for (WatchEvent<?> watchEvent : watchKey.pollEvents()) {
                    WatchEvent.Kind<?> kind = watchEvent.kind();
                    if (kind == StandardWatchEventKinds.OVERFLOW) {
                        continue;
                    }
                    // we sleep here a bit, to give time to the OS to complete file writing
                    Thread.sleep(1000L);
                    WatchEvent<Path> pathWatchEvent = (WatchEvent<Path>) watchEvent;
                    String watchEventContext = pathWatchEvent.context().toString();
                    Path p = path.resolve(Event.INCOMING).resolve(watchEventContext);
                    logger.log(Level.FINE, "watch event " + pathWatchEvent + " context = " + watchEventContext + " path = " + p);
                    if (pathEventManager.getSuspendedQueues().contains(name)) {
                        failEvent(p);
                    } else {
                        postEvent(p);
                    }
                }
                watchKey.reset();
            }
        } catch (ClosedWatchServiceException e) {
            logger.log(Level.SEVERE, "closed watch key: " + e.getMessage(), e);
        } catch (InterruptedException e) {
            logger.log(Level.WARNING, "interrupted");
            Thread.currentThread().interrupt();
        } catch (Exception e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
        }
        return eventCount;
    }

    @Override
    public void close() throws IOException {
        keepWatching = false;
        if (watchService != null) {
            logger.log(Level.FINE, "closing watch service " + watchService);
            watchService.close();
        }
    }

    public void drainIncoming() throws IOException {
        try (DirectoryStream<Path> stream = Files.newDirectoryStream(path.resolve(Event.INCOMING))) {
            stream.forEach(this::postEvent);
        }
    }

    public void purge() throws IOException {
        purge(path);
    }

    public void purge(Path path) throws IOException {
        try (DirectoryStream<Path> stream = Files.newDirectoryStream(path)) {
            stream.forEach(p -> {
                try {
                    if (Files.isRegularFile(p)) {
                        BasicFileAttributeView view = Files.getFileAttributeView(p, BasicFileAttributeView.class);
                        BasicFileAttributes attrs = view.readAttributes();
                        if (Instant.now().minusMillis(attrs.lastModifiedTime().toMillis()).toEpochMilli() > lifetime.millis()) {
                            logger.log(Level.WARNING, "lifetime of " + lifetime + " exceeded, deleting " + p);
                            Files.delete(p);
                        }
                    } else {
                        purge(p);
                    }
                } catch (IOException e) {
                    throw new UncheckedIOException(e);
                }
            });
        }
    }

    private void postEvent(Path file) {
        try {
            Event event = EventManager.eventFromFile(file);
            eventBus.post(event);
            eventCount++;
        } catch (IOException e) {
            logger.log(Level.SEVERE, "ignoring event post because of  " + e.getMessage());
        }
    }

    private void failEvent(Path file) {
        try {
            Event event = EventManager.eventFromFile(file);
            event.fail();
        } catch (IOException e) {
            logger.log(Level.SEVERE, "unable to fail event because of  " + e.getMessage());
        }
    }
}
