package org.xbib.event.timer;

import org.xbib.event.Payload;
import org.xbib.event.bus.EventBus;
import org.xbib.event.common.EventManager;
import org.xbib.event.common.EventManagerService;
import org.xbib.event.persistence.FilePersistenceStore;
import org.xbib.event.persistence.PersistenceStore;
import org.xbib.settings.Settings;
import org.xbib.time.chronic.Chronic;
import org.xbib.time.chronic.Span;

import java.io.IOException;
import java.text.ParseException;
import java.time.Instant;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

public class TimerEventManagerService implements EventManagerService {

    private static final Logger logger = Logger.getLogger(TimerEventManagerService.class.getName());

    private Map<String, TimerEventService> services;

    public TimerEventManagerService() {
    }

    @SuppressWarnings("unchecked")
    @Override
    public TimerEventManagerService init(EventManager eventManager) {
        Settings settings = eventManager.getSettings();
        EventBus eventBus = eventManager.getEventBus();
        this.services = new LinkedHashMap<>();
        for (Map.Entry<String, Settings> entry : settings.getGroups("event.timer").entrySet()) {
            String name = entry.getKey();
            Settings timerSettings = entry.getValue();
            try {
                PersistenceStore<String, Object> persistenceStore = new FilePersistenceStore(timerSettings, name);
                TimerEventService timerEventService = new TimerEventService(eventBus, name, ZoneId.systemDefault(), persistenceStore);
                services.put(name, timerEventService);
                logger.log(Level.INFO, "timer " + name + " active: " + timerEventService);
            } catch (Exception e) {
                logger.log(Level.WARNING, "unable to activate timer " + name + ", reason: " + e.getMessage(), e);
            }
        }
        return this;
    }

    public boolean publish(String service,
                           String timeSpec,
                           Payload payload) throws ParseException, IOException {
        if (services.containsKey(service)) {
            Span span = Chronic.parse(timeSpec);
            if (span != null) {
                ZonedDateTime zonedDateTime = span.getBeginCalendar();
                services.get(service).schedule(zonedDateTime.toInstant(), payload);
                logger.log(Level.INFO, "scheduled to " + zonedDateTime);
            } else {
                logger.log(Level.INFO, "timer event key " + service + ": can not understand time spec " + timeSpec);
                return false;
            }
            return true;
        } else {
            logger.log(Level.SEVERE, "unknown timer event key: " + service);
        }
        return false;
    }

    public boolean publish(String service,
                           Instant instant,
                           Payload payload) throws IOException {
        if (services.containsKey(service)) {
            services.get(service).schedule(instant, payload);
            return true;
        } else {
            logger.log(Level.SEVERE, "unknown timer event key: " + service);
        }
        return false;
    }

    public void purge() {
        for (Map.Entry<String, TimerEventService> entry : services.entrySet()) {
            logger.log(Level.INFO, "purging timer " + entry.getKey());
            entry.getValue().purge();
        }
    }

    @Override
    public void shutdown() throws IOException {
        for (Map.Entry<String, TimerEventService> entry : services.entrySet()) {
            logger.log(Level.INFO, "closing timer " + entry.getKey());
            entry.getValue().close();
        }
    }
}
